# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
"""
Package for authentication/identification of requests.

The objective of this package is to provide single-focused middleware
components that implement a particular specification.  Integration of
the components into a usable system is up to a higher-level framework.
"""
